#include "ds50daq/DAQ/CompositeDriver.hh"

#include "artdaq/Application/GeneratorMacros.hh"
#include "artdaq/Application/makeCommandableFragmentGenerator.hh"
#include "art/Utilities/Exception.h"
#include "cetlib/exception.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

using fhicl::ParameterSet;

ds50::CompositeDriver::CompositeDriver(ParameterSet const & ps):
  DS50FragmentGenerator(ps.get<ParameterSet>("generator_ds50"))
{
  std::vector<ParameterSet> psetList =
    ps.get<std::vector<ParameterSet>>("generator_config_list");
  for (auto pset : psetList) {
    if (! makeChildGenerator_(pset)) {
      throw cet::exception("CompositeDriver")
        << "Unable to create child generator for PSet= \""
        << pset.to_string() << "\"";
    }
  }
}

ds50::CompositeDriver::~CompositeDriver() noexcept
{
}

void ds50::CompositeDriver::start_()
{
  for (size_t idx = 0; idx < generator_active_list_.size(); ++idx) {
    generator_active_list_[idx] = true;
  }
  for (auto& generator : generator_list_) {
    generator->StartCmd(run_number());
  }
}

void ds50::CompositeDriver::stop_()
{
  std::vector<std::unique_ptr<DS50FragmentGenerator>>::reverse_iterator riter;
  for (riter = generator_list_.rbegin(); riter != generator_list_.rend(); ++riter) {
    (*riter)->StopCmd();
  }
}

void ds50::CompositeDriver::pause_()
{
  std::vector<std::unique_ptr<DS50FragmentGenerator>>::reverse_iterator riter;
  for (riter = generator_list_.rbegin(); riter != generator_list_.rend(); ++riter) {
    (*riter)->PauseCmd();
  }
}

void ds50::CompositeDriver::resume_()
{
  for (size_t idx = 0; idx < generator_active_list_.size(); ++idx) {
    generator_active_list_[idx] = true;
  }
  for (auto& generator : generator_list_) {
    generator->ResumeCmd();
  }
}

void ds50::CompositeDriver::shutdown_()
{
  std::vector<std::unique_ptr<DS50FragmentGenerator>>::reverse_iterator riter;
  for (riter = generator_list_.rbegin(); riter != generator_list_.rend(); ++riter) {
    (*riter)->shutdown();
  }
}

std::vector<artdaq::Fragment::fragment_id_t> ds50::CompositeDriver::fragmentIDs_()
{
  std::vector<artdaq::Fragment::fragment_id_t> workList;
  for (size_t idx = 0; idx < generator_list_.size(); ++idx) {
    std::vector<artdaq::Fragment::fragment_id_t> tempList =
      generator_list_[idx]->fragmentIDs();
    workList.insert(workList.end(), tempList.begin(), tempList.end());
  }     
  return workList;
}

bool ds50::CompositeDriver::getNext__(artdaq::FragmentPtrs& frags)
{
  bool anyGeneratorIsActive = false;
  for (size_t idx = 0; idx < generator_list_.size(); ++idx) {
    if (generator_active_list_[idx]) {
      bool status = generator_list_[idx]->getNext(frags);
      generator_active_list_[idx] = status;
      if (status) {anyGeneratorIsActive = true;}
    }
  }
  return anyGeneratorIsActive;
}

bool ds50::CompositeDriver::makeChildGenerator_(fhicl::ParameterSet const &pset)
{
  // pull out the relevant parts of the ParameterSet
  fhicl::ParameterSet daq_pset;
  try {
    daq_pset = pset.get<fhicl::ParameterSet>("daq");
  }
  catch (...) {
    mf::LogError("CompositeDriver::makeChildGenerator_")
      << "Unable to find the DAQ parameters in the initialization "
      << "ParameterSet: \"" + pset.to_string() + "\".";
    return false;
  }
  fhicl::ParameterSet fr_pset;
  try {
    fr_pset = daq_pset.get<fhicl::ParameterSet>("fragment_receiver");
  }
  catch (...) {
    mf::LogError("CompositeDriver::makeChildGenerator_")
      << "Unable to find the fragment_receiver parameters in the DAQ "
      << "initialization ParameterSet: \"" + daq_pset.to_string() + "\".";
    return false;
  }

  // create the requested FragmentGenerator
  std::string frag_gen_name = fr_pset.get<std::string>("generator", "");
  if (frag_gen_name.length() == 0) {
    mf::LogError("CompositeDriver::makeChildGenerator_")
      << "No fragment generator (parameter name = \"generator\") was "
      << "specified in the fragment_receiver ParameterSet.  The "
      << "DAQ initialization PSet was \"" << daq_pset.to_string() << "\".";
    return false;
  }

  std::unique_ptr<artdaq::CommandableFragmentGenerator> tmp_gen_ptr;
  try {
    tmp_gen_ptr = artdaq::makeCommandableFragmentGenerator(frag_gen_name, fr_pset);
  }
  catch (art::Exception& excpt) {
    mf::LogError("CompositeDriver::makeChildGenerator_")
      << "Exception creating a FragmentGenerator of type \""
      << frag_gen_name << "\" with parameter set \"" << fr_pset.to_string()
      << "\", exception = " << excpt;
    return false;
  }
  catch (cet::exception& excpt) {
    mf::LogError("CompositeDriver::makeChildGenerator_")
      << "Exception creating a FragmentGenerator of type \""
      << frag_gen_name << "\" with parameter set \"" << fr_pset.to_string()
      << "\", exception = " << excpt;
    return false;
  }
  catch (...) {
    mf::LogError("CompositeDriver::makeChildGenerator_")
      << "Unknown exception creating a FragmentGenerator of type \""
      << frag_gen_name << "\" with parameter set \"" << fr_pset.to_string()
      << "\".";
    return false;
  }

  std::unique_ptr<DS50FragmentGenerator> generator_ptr;
  generator_ptr.reset(nullptr);
  try {
    DS50FragmentGenerator* tmp_ds50gen_bareptr =
      dynamic_cast<DS50FragmentGenerator*>(tmp_gen_ptr.get());
    if (tmp_ds50gen_bareptr) {
      tmp_gen_ptr.release();
      generator_ptr.reset(tmp_ds50gen_bareptr);
    }
  }
  catch (...) {}
  if (! generator_ptr) {
    mf::LogError("CompositeDriver::makeChildGenerator_")
      << "Error: The requested fragment generator type (" << frag_gen_name
      << ") is not a DS50FragmentGenerator, and only DS50FragmentGenerators "
      << "are currently supported.";
    return false;
  }

  generator_list_.push_back(std::move(generator_ptr));
  generator_active_list_.push_back(true);
  return true;
}

DEFINE_ARTDAQ_COMMANDABLE_GENERATOR(ds50::CompositeDriver)
